/**
 * Copyright (c) 2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.LogRecord;

import javax.xml.namespace.QName;

import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.TraceCode;

/**
 * A MONIT log trace filter
 * 
 * @author Christophe DENEUX - Linagora
 *
 */
public class MonitLogFilter {

    private final List<LogRecord> monitLogs;

    public MonitLogFilter(final List<LogRecord> allMonitLogs) {
        this.monitLogs = new ArrayList<>(allMonitLogs);
    }

    /**
     * Filters MONIT log traces on the given trace code
     */
    public MonitLogFilter traceCode(final TraceCode expectedTraceCode) {

        return property(FlowLogData.FLOW_STEP_TRACE_CODE, expectedTraceCode);

    }

    /**
     * Filters MONIT log traces on the given interface
     */
    public MonitLogFilter interfaceName(final QName interfaceName) {

        return property(FlowLogData.FLOW_STEP_INTERFACE_NAME_PROPERTY_NAME, interfaceName.toString());

    }

    /**
     * Filters MONIT log traces on the given service
     */
    public MonitLogFilter serviceName(final QName serviceName) {

        return property(FlowLogData.FLOW_STEP_SERVICE_NAME_PROPERTY_NAME, serviceName.toString());

    }

    /**
     * Filters MONIT log traces on the given operation
     */
    public MonitLogFilter operationName(final QName operationName) {

        return property(FlowLogData.FLOW_STEP_OPERATION_NAME_PROPERTY_NAME, operationName.toString());

    }

    /**
     * Filters MONIT log traces on the given flow instance id
     */
    public MonitLogFilter flowInstanceId(final String flowInstanceId) {

        return property(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME, flowInstanceId);

    }

    /**
     * Filters MONIT log traces on the given property
     */
    public MonitLogFilter property(final String propertyName, final Object propertyValue) {

        return this.filterOn(new Filter() {

            @Override
            public boolean match(final FlowLogData flowLogData) {
                return propertyValue.equals(flowLogData.get(propertyName));
            }
        });

    }

    /**
     * @return MONIT log traces filtered. Only one trace is expected !
     */
    public LogRecord singleResult() {
        assertFalse("No existing log record remains!", this.monitLogs.isEmpty());
        assertEquals("Several log records exist again!", 1, this.monitLogs.size());
        return this.monitLogs.get(0);
    }

    /**
     * @return MONIT log traces filtered.
     */
    public List<LogRecord> results() {
        return new ArrayList<>(this.monitLogs);
    }

    private interface Filter {
        public boolean match(final FlowLogData flowLogData);
    }

    private MonitLogFilter filterOn(final Filter filter) {

        final Iterator<LogRecord> itMonitLogs = this.monitLogs.iterator();
        while (itMonitLogs.hasNext()) {
            final LogRecord currentLogRecord = itMonitLogs.next();

            final Object[] parameters = currentLogRecord.getParameters();
            assertNotNull("The log record does not contain parameters!", parameters);
            assertEquals("Unexpected number of parameters into the log record!", 1, parameters.length);
            final Object traceDetails = parameters[0];
            assertNotNull("The parameter of the log record does not contain trace details!", traceDetails);
            assertTrue("Unexpected trace detail object!", traceDetails instanceof FlowLogData);
            final FlowLogData flowLogData = (FlowLogData) traceDetails;
            if (!filter.match(flowLogData)) {
                itMonitLogs.remove();
            }
        }

        return this;
    }

}
