/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.jar.JarFile;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;
import javax.jbi.management.DeploymentException;
import javax.jbi.management.DeploymentServiceMBean;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.apache.commons.io.FileUtils;
import org.junit.Assert;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.DefaultBootstrap;
import org.ow2.petals.component.framework.JBIBootstrap;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.junit.impl.mock.MockComponentContext;
import org.ow2.petals.component.framework.junit.impl.mock.MockEndpointDirectory;
import org.ow2.petals.component.framework.junit.impl.mock.MockInstallationContext;
import org.ow2.petals.component.framework.junit.mbean.EmbeddedJmxServerConnector;
import org.ow2.petals.component.framework.process.JBIProcessorManagerObserver;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;

import com.ebmwebsourcing.easycommons.io.ZipHelper;
import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public class ComponentInstanceImpl {

    private enum ComponentState {
        NOT_INSTALLED, STOPPED, STARTED;
    }

    private static final String JBI_DIR_NAME = "jbi";

    private final File componentRepositoryRoot;

    private final ComponentConfiguration componentConfiguration;

    private final Jbi jbi;

    private final AbstractComponent component;

    private MockComponentContext mockComponentContext;

    private ComponentState componentState = ComponentState.NOT_INSTALLED;

    private InstallationContext mockInstallationContext;

    private JBIBootstrap componentBootstrap;

    private EmbeddedJmxServerConnector embeddedJmxSrv = null;

    public ComponentInstanceImpl(final MockEndpointDirectory directory,
            final ComponentConfiguration componentConfiguration, final File componentRepositoryRoot,
            final Logger componentLogger) {
        assert componentConfiguration != null;
        assert componentRepositoryRoot != null;

        this.componentRepositoryRoot = componentRepositoryRoot;
        System.out.println(String.format("------------- Created component under test, repository root '%s'",
                this.componentRepositoryRoot.getAbsolutePath()));
        this.componentConfiguration = componentConfiguration;
        this.jbi = generateComponentJbi();
        copyOthersJbiResources();
        copyJbiResources(componentConfiguration.getResources(), getInstallRoot());
        this.component = createComponent(jbi.getComponent().getComponentClassName().getContent());
        this.mockComponentContext = new MockComponentContext(directory, getComponentName(), this.component,
                this.getInstallRoot(), this.getWorkspaceRoot(), componentLogger);
    }

    public void setEmbeddedJmxSrv(final EmbeddedJmxServerConnector embeddedJmxSrv) {
        this.embeddedJmxSrv = embeddedJmxSrv;
    }

    private final File getInstallRoot() {
        return new File(componentRepositoryRoot, "install");
    }

    private final File getWorkspaceRoot() {
        return new File(componentRepositoryRoot, "work");
    }

    /**
     * Generate the JBI descriptor from the current component configuration
     * 
     * @return The JBI descriptor
     */
    private Jbi generateComponentJbi() {

        final File generatedJbiFile = this.componentConfiguration.generate(getInstallRoot());

        assert generatedJbiFile.exists();

        try (final FileInputStream fis = new FileInputStream(generatedJbiFile)) {
            return CDKJBIDescriptorBuilder.getInstance().buildJavaJBIDescriptor(fis);
        } catch (final IOException | JBIDescriptorException e) {
            throw new UncheckedException(e);
        }
    }

    public final String getComponentName() {
        return jbi.getComponent().getIdentification().getName();
    }

    public AbstractComponent getComponentObject() {
        return this.component;
    }

    private void copyOthersJbiResources() {

        final URL executionDirUrl = this.getClass().getClassLoader().getResource(JBI_DIR_NAME);
        assert executionDirUrl != null;

        try {
            if (executionDirUrl.getProtocol().equals("jar")) {

                final String executionDirStr = executionDirUrl.toString();
                final int executionJarEndIndex = executionDirStr.indexOf(".jar!") + ".jar".length();
                assert executionJarEndIndex != -1;

                final String executionJarStr = executionDirStr.substring("jar:file:".length(), executionJarEndIndex);
                final JarFile executionJar = new JarFile(executionJarStr);

                final File jbiDirExtractedFromComponentJar = new File(this.componentRepositoryRoot,
                        "jbiDirExtractedFromComponentJar");
                ZipHelper.extractEntriesFromArchive(executionJar, JBI_DIR_NAME + File.separator,
                        jbiDirExtractedFromComponentJar);
                this.copyAllJbiDirContentExceptJbiXmlFile(new File(jbiDirExtractedFromComponentJar, JBI_DIR_NAME));
            } else {
                final File executionDir = new File(executionDirUrl.toURI());
                this.copyAllJbiDirContentExceptJbiXmlFile(executionDir);
            }
        } catch (final Exception e) {
            throw new UncheckedException(e);
        }
    }

    private void copyAllJbiDirContentExceptJbiXmlFile(final File jbiDir) throws IOException {
        for (final File fileInJbiDir : jbiDir.listFiles()) {
            if ("jbi.xml".equals(fileInJbiDir.getName())) {
                // jbi.xml was generated just before, skip it.
                continue;
            }
            if (fileInJbiDir.isDirectory()) {
                FileUtils.copyDirectoryToDirectory(fileInJbiDir, this.getInstallRoot());
            } else {
                FileUtils.copyFileToDirectory(fileInJbiDir, this.getInstallRoot());
            }
        }
    }

    private void copyJbiResources(final List<URL> resourceUrls, final File installationRoot) {
        for (final URL resourceUrl : resourceUrls) {
            try {
                if (resourceUrl.getProtocol().equals("jar")) {
                    // TODO : not supported, to be implemented... otherwise
                    // resource copying when resource in jar will not work
                    continue;
                }
                final File resourceFile = new File(resourceUrl.toURI());
                if (resourceFile.isDirectory()) {
                    FileUtils.copyDirectory(resourceFile, installationRoot);
                } else {
                    FileUtils.copyFileToDirectory(resourceFile, installationRoot);
                }
            } catch (final Exception e) {
                throw new UncheckedException(e);
            }
        }
    }

    private void copyWsdl(final URL serviceDescription, final File installationRoot) {
        try {
            final File resourceFile = new File(serviceDescription.toURI());
            FileUtils.copyFileToDirectory(resourceFile, installationRoot);
        } catch (final IOException | URISyntaxException e) {
            throw new UncheckedException(e);
        }
    }

    private AbstractComponent createComponent(final String componentClassName) {
        try {
            return (AbstractComponent) Class.forName(componentClassName).newInstance();
        } catch (final Exception e) {
            throw new UncheckedException(e);
        }
    }

    public void load() {
        this.mockInstallationContext = new MockInstallationContext(this.getInstallRoot(), this.jbi,
                this.mockComponentContext, true);
        // jbi bootstrap
        try {
            @SuppressWarnings("unchecked")
            final Class<? extends JBIBootstrap> jbiBootstrapClass = (Class<? extends JBIBootstrap>) Class
                    .forName(this.jbi.getComponent().getBootstrapClassName());
            this.componentBootstrap = jbiBootstrapClass.newInstance();
            this.componentBootstrap.init(this.mockInstallationContext);
            if (this.embeddedJmxSrv != null && this.componentBootstrap instanceof DefaultBootstrap) {
                this.embeddedJmxSrv.registerConfigurationInstallerMBean((DefaultBootstrap) this.componentBootstrap);
            }
        } catch (final Exception e) {
            throw new UncheckedException(e);
        }
    }

    public void install() {
        // jbi initialize
        try {
            this.componentBootstrap.onInstall();
            this.component.init(this.mockInstallationContext.getContext());
            this.componentState = ComponentState.STOPPED;
        } catch (final JBIException jbie) {
            throw new UncheckedException(jbie);
        }
    }

    public boolean isInstalled() {
        return this.componentState == ComponentState.STOPPED || this.componentState == ComponentState.STARTED;
    }

    public void start() {
        try {
            this.component.start();
            this.componentState = ComponentState.STARTED;
        } catch (final JBIException e) {
            throw new UncheckedException("Impossible to start the wrapped component", e);
        }
    }

    public boolean isStarted() {
        return this.componentState == ComponentState.STARTED;
    }

    public void stop() {
        try {
            this.component.stop();
            this.componentState = ComponentState.STOPPED;
        } catch (final JBIException e) {
            throw new UncheckedException("Impossible to stop the component under test", e);
        }
    }

    public void shutdown() {
        try {
            if (this.isStarted()) {
                this.stop();
            }

            // To shutdown the component, all SUs must be undeployed
            final AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
            if (serviceUnitManager != null) {
                undeployAllServices();
            }
            this.component.shutDown();
            this.componentState = ComponentState.NOT_INSTALLED;
        } catch (final JBIException e) {
            throw new UncheckedException("Impossible to stop the component under test", e);
        }
    }

    public void deployService(final String suName, final ServiceConfiguration serviceConfiguration)
            throws DeploymentException {

        final AbstractServiceUnitManager serviceUnitManager = getSUManager();
        final File serviceUnitDir = this.getServiceConfigurationInstallDir(suName);

        if (serviceConfiguration instanceof ProvidesServiceConfiguration
                && ((ProvidesServiceConfiguration) serviceConfiguration).getServiceDescription() != null) {
            this.copyWsdl(((ProvidesServiceConfiguration) serviceConfiguration).getServiceDescription(),
                    serviceUnitDir);
        }
        this.copyJbiResources(serviceConfiguration.getResources(), serviceUnitDir);

        serviceConfiguration.generate(serviceUnitDir);

        final String result = this.deploySU(serviceUnitManager, suName, serviceUnitDir);
        if (result.contains("<task-result>SUCCESS</task-result>")) {
            this.initSU(serviceUnitManager, suName, serviceUnitDir);
            this.startSU(serviceUnitManager, suName);
        } else {
            throw new DeploymentException("Impossible to install the specified service: " + result);
        }
        final Collection<ServiceEndpoint> endpoints = serviceUnitManager.getSUDataHandler(suName).getServiceEndpoints();
        if (serviceConfiguration instanceof ProvidesServiceConfiguration) {
            final ProvidesServiceConfiguration scp = (ProvidesServiceConfiguration) serviceConfiguration;
            // if it was autogenerated, let's get the generated name
            if ("autogenerate".equals(scp.getEndpointName())) {
                scp.setEndpointName(endpoints.iterator().next().getEndpointName());
            }

            // we deployed only one provides!
            Assert.assertTrue(endpoints.size() == 1);
        } else {
            // we deployed no provides
            Assert.assertTrue(endpoints.isEmpty());
        }
    }

    private String deploySU(final AbstractServiceUnitManager serviceUnitManager, final String suName,
            final File serviceUnitDir) throws DeploymentException {
        try {
            return serviceUnitManager.deploy(suName, serviceUnitDir.getAbsolutePath());
        } catch (final DeploymentException e) {
            this.getLogger().log(Level.SEVERE, String.format("The deployment of service unit '%s' fails.", suName), e);
            throw e;
        }
    }

    private void initSU(final AbstractServiceUnitManager serviceUnitManager, final String suName,
            final File serviceUnitDir) throws DeploymentException {
        try {
            serviceUnitManager.init(suName, serviceUnitDir.getAbsolutePath());
        } catch (final DeploymentException e) {
            this.getLogger().log(Level.SEVERE, String.format("The initialization of service unit '%s' fails.", suName),
                    e);
            throw e;
        }
    }

    private void startSU(final AbstractServiceUnitManager serviceUnitManager, final String suName)
            throws DeploymentException {
        try {
            serviceUnitManager.start(suName);
        } catch (final DeploymentException e) {
            this.getLogger().log(Level.SEVERE, String.format("The start-up of service unit '%s' fails.", suName), e);
            throw e;
        }
    }

    private AbstractServiceUnitManager getSUManager() {
        final AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
        Assert.assertNotNull("It makes no sense to deploy a service on a component that does not support SU deployment",
                serviceUnitManager);
        return serviceUnitManager;
    }

    public void initService(final String suName) throws DeploymentException {
        final File serviceUnitDir = getServiceConfigurationInstallDir(suName);
        getSUManager().init(suName, serviceUnitDir.getAbsolutePath());
    }

    public void startService(final String suName) throws DeploymentException {
        getSUManager().start(suName);
    }

    public void stopService(final String suName) throws DeploymentException {
        getSUManager().stop(suName);
    }

    public void shutdownService(final String suName) throws DeploymentException {
        getSUManager().shutDown(suName);
    }

    public File getServiceConfigurationInstallDir(final String suName) {
        final File serviceUnitsInstallRoot = new File(this.componentRepositoryRoot, "service-units");
        return new File(serviceUnitsInstallRoot, suName);
    }

    public void undeployService(final String suName) {
		try {
            final AbstractServiceUnitManager serviceUnitManager = getSUManager();

            final ServiceUnitDataHandler serviceUnitDataHandler = serviceUnitManager.getSUDataHandler(suName);
			serviceUnitManager.stop(suName);
			serviceUnitManager.shutDown(suName);
			serviceUnitManager.undeploy(suName,	serviceUnitDataHandler.getInstallRoot());
        } catch (final DeploymentException e) {
			throw new UncheckedException(
					"Impossible to uninstall the specified service", e);
		}
    }

	public void undeployAllServices() {
		try {
            final AbstractServiceUnitManager serviceUnitManager = getSUManager();
	        
            for (ServiceUnitDataHandler serviceUnitDataHandler : serviceUnitManager.getServiceUnitDataHandlers()) {
                final String suName = serviceUnitDataHandler.getName();
                if (serviceUnitDataHandler.getState().equals(DeploymentServiceMBean.STARTED)) {
                    serviceUnitManager.stop(suName);
                    Assert.assertEquals(DeploymentServiceMBean.STOPPED, serviceUnitDataHandler.getState());
                }

                if (serviceUnitDataHandler.getState().equals(DeploymentServiceMBean.STOPPED)) {
                    serviceUnitManager.shutDown(suName);
                    Assert.assertEquals(DeploymentServiceMBean.SHUTDOWN, serviceUnitDataHandler.getState());
                }
				serviceUnitManager.undeploy(suName,	serviceUnitDataHandler.getInstallRoot());
	        }
        } catch (final DeploymentException e) {
			throw new UncheckedException(
					"Impossible to uninstall all the services", e);
		}
	}
	
    public boolean isServiceDeployed(final String suName) {
        return getSUManager().getSUDataHandler(suName) != null;
    }

    public void pushMessageInDeliveryChannel(final PetalsMessageExchange msgExchange, final boolean sync) {
        final Semaphore sem;
        final JBIProcessorManagerObserver obs;
        if (sync) {
            sem = new Semaphore(0);
            obs = new JBIProcessorManagerObserver() {
                @Override
                public void beforeProcessingMessage(final Exchange exchange) {
                    // do nothing
                }

                @Override
                public void afterProcessingMessage(final Exchange exchange) {
                    if (exchange.getExchangeId().equals(msgExchange.getExchangeId())) {
                        // this will unblock the execution of pushSync until
                        // the message has been processed by the provider
                        sem.release();
                    }
                }
            };

            this.component.getProcessorManager().addObserver(obs);
        } else {
            sem = null;
            obs = null;
        }

        try {
            this.mockComponentContext.getDeliveryChannel().enqueueIncomingExchange(msgExchange);

            if (sync) {
                assert sem != null;
                if (!sem.tryAcquire(2, TimeUnit.SECONDS)) {
                    Assert.fail("The message delivery has timed out!");
                }
            }
        } catch (final InterruptedException e) {
            throw new UncheckedException("Impossible to synchronously send a message to the component", e);
        } finally {
            if (obs != null) {
                this.component.getProcessorManager().deleteObserver(obs);
            }
        }
    }

    public PetalsMessageExchange pollFromProvider(final long timeout) throws TimeoutException {
        return this.mockComponentContext.getDeliveryChannel().dequeueExchangeToClient(timeout);
    }

    public void clearResponsesFromProvider() {
        this.mockComponentContext.getDeliveryChannel().clearExchangesToClient();
    }

    public int getResponsesFromProviderCount() {
        return this.mockComponentContext.getDeliveryChannel().getExchangesToClientCount();
    }

    public PetalsMessageExchange pollFromConsumer(final long timeout) throws TimeoutException {
        return this.mockComponentContext.getDeliveryChannel().dequeueExchangeToServer(timeout);
    }

    public void clearRequestsFromConsumer() {
        this.mockComponentContext.getDeliveryChannel().clearExchangesToServer();
    }

    public int getRequestsFromConsumerCount() {
        return this.mockComponentContext.getDeliveryChannel().getExchangesToServerCount();
    }

    public void clearIncomingMessages() {
        this.mockComponentContext.getDeliveryChannel().clearIncomingExchanges();
    }

    public int getExchangesInDeliveryChannelCount() {
        return this.mockComponentContext.getDeliveryChannel().getExchangesToComponentCount();
    }

    public String getNativeEndpointName(final QName service) {
        final Description nativeDescription = this.component.getNativeWsdl().getDescription();
        return nativeDescription.getService(service).getEndpoints().get(0).getName();
    }

    public Logger getLogger() {
        return this.component.getLogger();
    }
}
