/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.ow2.petals.component.framework.junit.JbiConstants;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.xml.Transformers;

/**
 * @author Marc Jambert - EBM WebSourcing
 * @author Victor Noel
 */
public abstract class Configuration implements JbiConstants {

    private final Map<QName, String> parameters;

    private final List<URL> resourceUrls;

    public Configuration() {
        this.parameters = new HashMap<QName, String>();
        this.resourceUrls = new ArrayList<URL>();
    }

    public Map<QName, String> getParameters() {
        return Collections.unmodifiableMap(this.parameters);
    }

    public boolean hasParameter(final QName parameterName) {
        assert parameterName != null;
        return parameters.containsKey(parameterName);
    }

    public String getParameter(final QName parameterName) {
        assert parameterName != null;
        return parameters.get(parameterName);
    }

    public void setParameter(final QName parameterName, final String parameterValue) {
        assert parameterName != null;
        assert parameterValue != null;
        parameters.put(parameterName, parameterValue);
    }

    public void addResource(final URL resourceUrl) {
        assert resourceUrl != null;
        resourceUrls.add(resourceUrl);
    }

    public List<URL> getResources() {
        return Collections.unmodifiableList(resourceUrls);
    }

    protected static Element getOrCreateJbiElement(final Document document) {
        final Element jbiElement = addOrReplaceElement(document, document, new QName(JBI_NAMESPACE_URI, "jbi"));
        jbiElement.setAttribute("version", "1.0");
        jbiElement.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                XMLConstants.XMLNS_ATTRIBUTE + ":" + XSI_NAMESPACE_PREFIX, XMLConstants.W3C_XML_SCHEMA_INSTANCE_NS_URI);
        jbiElement.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                XMLConstants.XMLNS_ATTRIBUTE + ":" + JBI_NAMESPACE_PREFIX, JBI_NAMESPACE_URI);
        jbiElement.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                XMLConstants.XMLNS_ATTRIBUTE + ":" + CDK_NAMESPACE_PREFIX, CDK_NAMESPACE_URI);
        return jbiElement;
    }

    protected static Element getJbiElement(final Document document) {
        return getElementByTagNameNS(document, JbiConstants.JBI_NAMESPACE_URI, "jbi");
    }

    protected static Element addElement(final Document jbiDocument, final Node container, final QName parameterQName) {
        final Element parameter = jbiDocument.createElementNS(parameterQName.getNamespaceURI(),
                parameterQName.getLocalPart());
        container.appendChild(parameter);

        return parameter;
    }

    protected static Element addOrReplaceElement(final Document jbiDocument, final Node container,
            final QName parameterQName) {
        return addOrReplaceElement(jbiDocument, container, parameterQName, null);
    }

    protected static Element addElement(final Document jbiDocument, final Node container, final QName parameterQName,
            final String parameterValue) {
        final Element parameter = jbiDocument.createElementNS(parameterQName.getNamespaceURI(),
                parameterQName.getLocalPart());

        container.appendChild(parameter);

        if (parameterValue != null) {
            parameter.setTextContent(parameterValue);
        }

        return parameter;
    }

    protected static Element addOrReplaceElement(final Document jbiDocument, final Node container,
            final QName parameterQName, final String parameterValue) {
        assert container != null;
        assert parameterQName != null;

        Element parameter = getElementByTagNameNS(container, parameterQName.getNamespaceURI(),
                parameterQName.getLocalPart());

        if (parameter == null) {
            parameter = addElement(jbiDocument, container, parameterQName);
        }

        if (parameterValue != null) {
            parameter.setTextContent(parameterValue);
        }

        return parameter;
    }

    protected static Element getElementByTagNameNS(final Node container, final String namespaceURI,
            final String localName) {
        // we only want the children of container, not all the tree...
        final NodeList nodes = container.getChildNodes();
        Element res = null;
        for (int i = 0; i < nodes.getLength(); i++) {
            final Node item = nodes.item(i);
            if (item instanceof Element && namespaceURI.equals(item.getNamespaceURI())
                    && localName.equals(item.getLocalName())) {
                if (res != null) {
                    throw new UncheckedException(String.format("More than one element for {%s}%s in %s", namespaceURI,
                            localName, container));
                }
                res = (Element) item;
            }
        }
        return res;
    }

    /**
     * Override to add extra custom configuration on the JBI document (such as adding elements more complex than simple
     * parameters...).
     */
    protected void extraJBIConfiguration(final Document jbiDocument) {
        // nothing
    }

    /**
     * Meant to be overriden only by {@link ComponentConfiguration} and {@link ServiceConfiguration}.
     * 
     * Hence it has package visibility.
     */
    abstract Document initialiseJBIDocument(final File installationRoot);

    /**
     * Meant to be called only by {@link ComponentInstanceImpl}.
     * 
     * Hence it has package visibility.
     */
    File generate(final File installationRoot) {
        assert installationRoot != null;

        final Document jbiDocument = initialiseJBIDocument(installationRoot);

        extraJBIConfiguration(jbiDocument);

        // dump JBI descriptor back to installationRoot.
        final File installationRootMetaInf = new File(installationRoot, "META-INF");
        installationRootMetaInf.mkdirs();
        final File jbiDumpTarget = new File(installationRootMetaInf, "jbi.xml");
        final Transformer transformer = Transformers.takeTransformer();
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
        try (final OutputStream fos = new FileOutputStream(jbiDumpTarget)) {
            transformer.transform(new DOMSource(jbiDocument), new StreamResult(fos));
            return jbiDumpTarget;
        } catch (final IOException | TransformerException e) {
            throw new UncheckedException(e);
        } finally {
            Transformers.releaseTransformer(transformer);
        }
    }
}
