/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl;

import java.io.File;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;

/**
 * 
 * The justification behind having this strange architecture where services defines their dependencies as well as
 * parameters in the services node is to ease the quick definition of services in tests.
 * 
 * @author Marc Jambert - EBM WebSourcing
 * @author Victor Noel
 */
public class ServiceConfiguration extends Configuration {

    private static final String SERVICE_NAMESPACE_PREFIX = "srvNs";

    private static final String INTERFACE_NAMESPACE_PREFIX = "itfNs";
    
    private static final String OPERATION_NAMESPACE_PREFIX = "opNs";
    
    private final Map<QName, String> servicesSectionParameters = new HashMap<>();

    private final List<PCServiceConfiguration> serviceConfigurationDependencies = new ArrayList<>();

    public void addServiceConfigurationDependency(final PCServiceConfiguration serviceConfiguration) {
        assert serviceConfiguration != null;
        this.serviceConfigurationDependencies.add(serviceConfiguration);
    }

    /**
     * @return a copy of service dependencies {@link List}
     */
    public List<PCServiceConfiguration> getServiceConfigurationDependencies() {
        return Collections.unmodifiableList(this.serviceConfigurationDependencies);
    }

    /**
     * Parameters for the services section of the SU.
     * 
     */
    public Map<QName, String> getServicesSectionParameters() {
        return Collections.unmodifiableMap(this.servicesSectionParameters);
    }

    public boolean hasServicesSectionParameter(final QName parameterName) {
        assert parameterName != null;
        return servicesSectionParameters.containsKey(parameterName);
    }

    public String getServicesSectionParameter(final QName parameterName) {
        assert parameterName != null;
        return servicesSectionParameters.get(parameterName);
    }

    public void setServicesSectionParameter(final QName parameterName, final String parameterValue) {
        assert parameterName != null;
        assert parameterValue != null;
        servicesSectionParameters.put(parameterName, parameterValue);
    }

    protected static Element getOrCreateServicesElement(final Document document) {
        final Element jbi = getOrCreateJbiElement(document);
        return addOrReplaceElement(document, jbi, new QName(JBI_NAMESPACE_URI, "services"));
    }

    @Override
    Document initialiseJBIDocument(final File installationRoot) {

        final Document document = DocumentBuilders.newDocument();
        document.setXmlVersion("1.0");
        document.setXmlStandalone(true);

        final Element servicesElement = getOrCreateServicesElement(document);
        // this is not used but needed by the schema...
        servicesElement.setAttribute("binding-component", "true");

        if (this instanceof PCServiceConfiguration) {
            generateServiceTypeElement(document, servicesElement, this);
        }

        for (final ServiceConfiguration s : serviceConfigurationDependencies) {
            generateServiceTypeElement(document, servicesElement, s);
        }

        for (final Entry<QName, String> entry : servicesSectionParameters.entrySet()) {
            addOrReplaceElement(document, servicesElement, entry.getKey(), entry.getValue());
        }

        return document;
    }

    /**
     * Generate provide or consume element
     */
    private static void generateServiceTypeElement(final Document document, final Element servicesElement,
            final ServiceConfiguration serviceConfiguration) {
        
        final String serviceTypeElementName;
        if (serviceConfiguration instanceof ConsumesServiceConfiguration) {
            serviceTypeElementName = "consumes";
        } else if (serviceConfiguration instanceof ProvidesServiceConfiguration) {
            serviceTypeElementName = "provides";
        } else {
            throw new RuntimeException("Impossible case");
        }
        
        final PCServiceConfiguration sc = (PCServiceConfiguration) serviceConfiguration;

        final Element serviceElement = addOrReplaceElement(document, servicesElement,
                new QName(JBI_NAMESPACE_URI, serviceTypeElementName));

        // Create attributes
        if (sc.getEndpointName() != null) {
            serviceElement.setAttribute("endpoint-name", sc.getEndpointName());
        }

        serviceElement.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                XMLConstants.XMLNS_ATTRIBUTE + ":" + INTERFACE_NAMESPACE_PREFIX,
                sc.getInterfaceName().getNamespaceURI());
        serviceElement.setAttribute("interface-name",
                INTERFACE_NAMESPACE_PREFIX + ":" + sc.getInterfaceName().getLocalPart());

        if (sc.getServiceName() != null) {
            serviceElement.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                    XMLConstants.XMLNS_ATTRIBUTE + ":" + SERVICE_NAMESPACE_PREFIX,
                    sc.getServiceName().getNamespaceURI());
            serviceElement.setAttribute("service-name",
                    SERVICE_NAMESPACE_PREFIX + ":" + sc.getServiceName().getLocalPart());
        }

        final Long timeout = sc.getTimeout();
        if (timeout != null) {
            addOrReplaceElement(document, serviceElement, new QName(CDK_NAMESPACE_URI, "timeout"),
                    Long.toString(timeout));
        }

        if (serviceConfiguration instanceof ConsumesServiceConfiguration) {
            final ConsumesServiceConfiguration scc = (ConsumesServiceConfiguration) serviceConfiguration;
            final QName operation = scc.getOperation();
            if (operation != null) {
                final Element opElement = addOrReplaceElement(document, serviceElement,
                        new QName(CDK_NAMESPACE_URI, "operation"),
                        OPERATION_NAMESPACE_PREFIX + ":" + operation.getLocalPart());
                opElement.setAttributeNS(XMLConstants.XMLNS_ATTRIBUTE_NS_URI,
                        XMLConstants.XMLNS_ATTRIBUTE + ":" + OPERATION_NAMESPACE_PREFIX,
                        operation.getNamespaceURI());
            }

            final MEPType mep = scc.getMEP();
            if (mep != null) {
                addOrReplaceElement(document, serviceElement, new QName(CDK_NAMESPACE_URI, "mep"), mep.value());
            } else {
                final Element e = addOrReplaceElement(document, serviceElement, new QName(CDK_NAMESPACE_URI, "mep"));
                // shortcomming of jbi cdk xsd: mep is required...
                e.setAttributeNS(XMLConstants.W3C_XML_SCHEMA_INSTANCE_NS_URI, XSI_NAMESPACE_PREFIX + ":nil", "true");
            }
        }

        if (serviceConfiguration instanceof ProvidesServiceConfiguration) {
            final ProvidesServiceConfiguration scp = (ProvidesServiceConfiguration) serviceConfiguration;
            final Element wsdlEl = addOrReplaceElement(document, serviceElement, new QName(CDK_NAMESPACE_URI, "wsdl"));
            if (scp.getServiceDescription() != null) {
                try {
                    final File wsdl = new File(scp.getServiceDescription().toURI());
                    wsdlEl.setTextContent(wsdl.getName());
                } catch (final URISyntaxException e) {
                    throw new UncheckedException(e);
                }
            }
        }

        // append service parameters
        for (final Entry<QName, String> entry : serviceConfiguration.getParameters().entrySet()) {
            addOrReplaceElement(document, serviceElement, entry.getKey(), entry.getValue());
        }

        sc.extraServiceConfiguration(document, serviceElement);
    }
}
