/**
 * Copyright (c) 2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.mbean;

import java.io.File;
import java.io.FileOutputStream;
import java.lang.management.ManagementFactory;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;
import javax.management.RuntimeOperationsException;
import javax.management.modelmbean.InvalidTargetObjectTypeException;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.mina.util.AvailablePortFinder;
import org.junit.rules.ExternalResource;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.component.framework.DefaultBootstrap;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.mbean.ExtensionMBean;

/**
 * The EmbeddedJmxServerConnector Rule allows creation of an embedded JMX server connector to use component MBeans that
 * are guaranteed to be deleted when the test method finishes (whether it passes or fails):
 * 
 * <pre>
 * public static class HasJmxServerConnector {
 *  &#064;Rule
 *  public EmbeddedJmxServerConnector jmxSrvCon = new EmbeddedJmxServerConnector(credentials);
 * 
 *  &#064;Test
 *  public void test...() throws IOException {
 *      jmxSrvCon.registerConfigurationInstallerMBean(...);
 *      ...
 *  }
 * }
 * </pre>
 */
public class EmbeddedJmxServerConnector extends ExternalResource {

    /**
     * A user having a non-empty password
     */
    protected static final String JMX_USER = "petals";

    protected static final String JMX_PASSWORD = "petals";

    protected static final String COMPONENT_ID = "component-under-test";

    /**
     * The {@link JMXConnectorServer} listen on the localhost
     */
    private static final String JMX_HOST = "localhost";

    /**
     * The listening port of the {@link JMXConnectorServer}
     */
    private final int jmxPort;

    /**
     * Credentials of the {@link JMXConnectorServer}
     */
    private final String[][] credentials;

    /**
     * Temporary folder required to store password and access files needed to the {@link JMXConnectorServer}
     */
    private final TemporaryFolder tempFolder = new TemporaryFolder();

    /**
     * The JMX server of the current Java platform
     */
    private final MBeanServer mbeanServer = ManagementFactory.getPlatformMBeanServer();

    /**
     * The JMX connector server
     */
    private JMXConnectorServer jmxConnectorServer;

    /**
     * List of the registered MBeans
     */
    private List<ObjectName> registeredMBeans = new ArrayList<>();

    /**
     * The JMX URL of the JMX connector server
     */
    private String jmxUrl;

    /**
     * Creates an EmbeddedJmxServerConnector.
     * 
     * @param credentials
     *            Credentials to secured the {@link JMXConnectorServer}
     */
    public EmbeddedJmxServerConnector(final String[][] credentials) {
        this.jmxPort = AvailablePortFinder.getNextAvailable(7700);
        this.credentials = credentials;
        this.jmxUrl = "service:jmx:rmi:///jndi/rmi://" + JMX_HOST + ":" + this.jmxPort + "/" + "PetalsCDKJMX";
    }

    /**
     * Creates an EmbeddedJmxServerConnector using the default credentials (petals/petals).
     */
    public EmbeddedJmxServerConnector() {
        this(new String[][] { { JMX_USER, JMX_PASSWORD } });
    }

    @Override
    protected void before() throws Throwable {
        this.startJmxServerConnector();
    }

    @Override
    protected void after() {
        try {
            this.stopJmxServerConnector();
        } catch (final Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Create and start a {@link JMXConnectorServer}
     */
    private void startJmxServerConnector() throws Exception {

        final Registry registry = LocateRegistry.getRegistry(this.jmxPort);
        try {
            registry.list();
        } catch (final Exception e) {
            LocateRegistry.createRegistry(this.jmxPort);
        }

        this.tempFolder.create();

        final HashMap<String, Object> environment;
        if (this.credentials != null && this.credentials.length > 0) {
            final File passwordFile = this.tempFolder.newFile("password.properties");
            final File accessFile = this.tempFolder.newFile("access.properties");

            final Properties pwdProps = new Properties();
            for (final String[] credential : this.credentials) {
                pwdProps.setProperty(credential[0], credential[1] == null ? "" : credential[1]);
            }
            final FileOutputStream fosPwd = new FileOutputStream(passwordFile);
            try {
                pwdProps.store(fosPwd, "Credentials file");
            } finally {
                fosPwd.close();
            }

            final Properties accessProps = new Properties();
            for (final String[] credential : this.credentials) {
                accessProps.setProperty(credential[0], "readwrite");
            }
            final FileOutputStream fosAccess = new FileOutputStream(accessFile);
            try {
                accessProps.store(fosAccess, "Access file");
            } finally {
                fosAccess.close();
            }

            environment = new HashMap<>();
            environment.put("jmx.remote.x.password.file", passwordFile.getAbsolutePath());
            environment.put("jmx.remote.x.access.file", accessFile.getAbsolutePath());
        } else {
            environment = null;
        }

        this.jmxConnectorServer = JMXConnectorServerFactory.newJMXConnectorServer(new JMXServiceURL(this.jmxUrl),
                environment, this.mbeanServer);
        this.jmxConnectorServer.start();

        this.registeredMBeans.clear();

    }

    /**
     * Stop and destroy the {@link JMXConnectorServer}
     */
    private void stopJmxServerConnector() throws Exception {
        for (final ObjectName registeredMbean : this.registeredMBeans) {
            this.mbeanServer.unregisterMBean(registeredMbean);
        }
        this.jmxConnectorServer.stop();
        this.tempFolder.delete();
    }

    public String getJmxUrl() {
        return this.jmxUrl;
    }

    /**
     * Register a configuration installation mbean. If a MBean is already registered, it is unregistered before
     * registration of the new one.
     * 
     * @param bootstrap
     *            The boostrap instance used by the configuration installation mbean to register.
     * @throws NotCompliantMBeanException
     * @throws RuntimeOperationsException
     * @throws InstanceNotFoundException
     * @throws MBeanException
     * @throws InvalidTargetObjectTypeException
     * @throws InstanceAlreadyExistsException
     *             The MBean was already registered, the unregistration succeeded but the new registration detects that
     *             the MBean exists always.
     * @throws MalformedObjectNameException
     * @throws PEtALSCDKException
     */
    public EmbeddedJmxServerConnector registerConfigurationInstallerMBean(final DefaultBootstrap bootstrap)
            throws InstanceAlreadyExistsException, NotCompliantMBeanException, RuntimeOperationsException,
            InstanceNotFoundException, MBeanException, InvalidTargetObjectTypeException, MalformedObjectNameException,
            PEtALSCDKException {

        this.registerService(new ObjectName("org.ow2.petals:type=custom,name=bootstrap_" + COMPONENT_ID),
                new ExtensionMBean(bootstrap));

        return this;
    }

    /**
     * Register a MBean in the JMX server
     * 
     * @param objName
     *            The MBean name.
     * @param mbean
     *            The MBean to register.
     * @throws NotCompliantMBeanException
     * @throws RuntimeOperationsException
     * @throws InstanceNotFoundException
     * @throws MBeanException
     * @throws InvalidTargetObjectTypeException
     * @throws InstanceAlreadyExistsException
     *             The MBean was already registered, the unregistration succeeded but the new registration detects that
     *             the MBean exists always.
     */
    private void registerService(final ObjectName objName, final Object mbean) throws MBeanRegistrationException,
            InstanceNotFoundException, NotCompliantMBeanException, InstanceAlreadyExistsException {
        try {
            this.mbeanServer.registerMBean(mbean, objName);
            this.registeredMBeans.add(objName);
        } catch (final InstanceAlreadyExistsException e) {
            this.mbeanServer.unregisterMBean(objName);
            this.mbeanServer.registerMBean(mbean, objName);
        }
    }

}
