/**
 * Copyright (c) 2015-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.helpers;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;

import javax.jbi.messaging.ExchangeStatus;

import org.custommonkey.xmlunit.Diff;
import org.hamcrest.Matcher;
import org.ow2.petals.component.framework.junit.Message;
import org.ow2.petals.component.framework.junit.rule.ComponentUnderTest;

import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * Abstract class and many implementations for various checks to do on {@link Message}s objects send and received
 * to/from a {@link ComponentUnderTest} .
 * 
 * It is exploited by {@link SimpleComponent}.
 * 
 * @author Victor Noel
 *
 */
public abstract class MessageChecks {

    /**
     * Checks to apply on a message.
     * 
     * @param message
     * @throws Exception
     */
    public abstract void checks(final Message message) throws Exception;

    public final MessageChecks andThen(final MessageChecks checks) {
        final MessageChecks me = this;
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                me.checks(message);
                checks.checks(message);
            }
        };
    }

    public static MessageChecks none() {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
            }
        };
    }

    public static MessageChecks matches(final Matcher<Message> matcher) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertThat(message, matcher);
            }
        };
    }

    public static MessageChecks hasFault() {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertNotNull(message.getFault());
            }
        };
    }

    public static MessageChecks noFault() {
        return new MessageChecks() {
            @Override
            public void checks(Message message) throws Exception {
                assertNull(message.getFault());
            }
        };
    }

    public static MessageChecks hasOut() {
        return new MessageChecks() {
            @Override
            public void checks(Message message) throws Exception {
                assertNotNull(message.getOut());
            }
        };
    }

    public static MessageChecks noOut() {
        return new MessageChecks() {
            @Override
            public void checks(Message message) throws Exception {
                assertNull(message.getOut());
            }
        };
    }

    public static MessageChecks hasError() {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertNotNull(message.getError());
            }
        };
    }

    public static MessageChecks noError() {
        return new MessageChecks() {
            @Override
            public void checks(Message message) throws Exception {
                assertNull(message.getMessageExchange().getError());
            }
        };
    }

    public static MessageChecks status(final ExchangeStatus status) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertEquals(status, message.getStatus());
            }
        };
    }

    public static MessageChecks isDone() {
        return status(ExchangeStatus.DONE);
    }

    public static MessageChecks onlyDone() {
        return MessageChecks.noFault().andThen(MessageChecks.noError()).andThen(MessageChecks.noOut())
                .andThen(MessageChecks.isDone());
    }

    public static MessageChecks isError() {
        return status(ExchangeStatus.ERROR);
    }

    public static MessageChecks isActive() {
        return status(ExchangeStatus.ACTIVE);
    }

    public static MessageChecks propertyNotExists(final String property) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertNull(message.getMessageExchange().getProperty(property));
            }
        };
    }

    public static MessageChecks propertyExists(final String property) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertNotNull(message.getMessageExchange().getProperty(property));
            }
        };
    }

    public static MessageChecks propertyEquals(final String property, final Object value) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertEquals(message.getMessageExchange().getProperty(property), value);
            }
        };
    }

    public static MessageChecks hasXmlContent(final String expectedContent) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                final Diff diff = new Diff(expectedContent, SourceHelper.toString(message.getPayload()));
                assertTrue(diff.similar());
            }
        };
    }

    public static MessageChecks contentContains(final String expectedSubContent) {
        return new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertTrue(SourceHelper.toString(message.getPayload()).contains(expectedSubContent));
            }
        };
    }

    public static MessageChecks errorContains(final String expectedSubContent) {
        return isError().andThen(new MessageChecks() {
            @Override
            public void checks(final Message message) throws Exception {
                assertTrue(message.getError().getMessage().contains(expectedSubContent));
            }
        });
    }
}