/**
 * Copyright (c) 2015-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.helpers;

import javax.jbi.messaging.ExchangeStatus;

import org.junit.Assert;
import org.ow2.petals.component.framework.junit.Message;
import org.ow2.petals.component.framework.junit.RequestMessage;
import org.ow2.petals.component.framework.junit.ResponseMessage;
import org.ow2.petals.component.framework.junit.StatusMessage;
import org.ow2.petals.component.framework.junit.impl.message.FaultToConsumerMessage;
import org.ow2.petals.component.framework.junit.impl.message.ResponseToConsumerMessage;
import org.ow2.petals.component.framework.junit.impl.message.StatusToConsumerMessage;
import org.ow2.petals.component.framework.junit.rule.ComponentUnderTest;


/**
 * 
 * Abstract class and some default implementation to define an external service a {@link ComponentUnderTest} could
 * interact with.
 * 
 * It is exploited by {@link SimpleComponent}.
 * 
 * @author Victor Noel
 *
 */
public abstract class ServiceProviderImplementation {

    /**
     * The implementation of an external provider service.
     * 
     * Can return {@link ResponseMessage} and {@link #statusExpected()} should return <code>true</code> ; or
     * {@link StatusMessage} and {@link #statusExpected()} should return false!
     */
    public abstract Message provides(final RequestMessage request) throws Exception;

    /**
     * This is only called by {@link SimpleComponent} if {@link #statusExpected} is <code>true</code>.
     * 
     * By default it will be thus called and fail to force its implementation!
     */
    public void handleStatus(final StatusMessage status) throws Exception {
        Assert.fail("Either define handleStatus() or set statusExpected to false.");
    }

    public boolean statusExpected() {
        return true;
    }

    /**
     * Checks executed before executing the consumer implementation (it is not for acknowledgement for the request
     * message message)
     */
    public final ServiceProviderImplementation with(final MessageChecks checks) {
        final ServiceProviderImplementation me = this;
        return new ServiceProviderImplementation() {
            @Override
            public Message provides(final RequestMessage request) throws Exception {
                checks.checks(request);
                return me.provides(request);
            }

            @Override
            public void handleStatus(final StatusMessage status) throws Exception {
                me.handleStatus(status);
            }

            @Override
            public boolean statusExpected() {
                return me.statusExpected();
            }
        };
    }

    /**
     * A service that returns a ResponseMessage with the given content as out message and the given checks for the
     * acknowledgement if there is one.
     * 
     * Set statusChecks to <code>null</code> if no final status is expected.
     */
    public static ServiceProviderImplementation outMessage(final String content, final MessageChecks statusChecks) {
        return new ServiceProviderImplementation() {
            @Override
            public Message provides(final RequestMessage request) throws Exception {
                return new ResponseToConsumerMessage(request, content);
            }

            @Override
            public void handleStatus(final StatusMessage status) throws Exception {
                assert statusChecks != null;
                statusChecks.checks(status);
            }

            @Override
            public boolean statusExpected() {
                return statusChecks != null;
            }
        };
    }

    /**
     * Like {@link #outMessage(String, MessageChecks)} with {@link ExchangeStatus#DONE} as expected final status.
     */
    public static ServiceProviderImplementation outMessage(final String content) {
        return outMessage(content, MessageChecks.status(ExchangeStatus.DONE));
    }

    /**
     * A service that returns a ResponseMessage with the given content as fault and the given checks for the
     * acknowledgement if there is one.
     * 
     * Set statusChecks to <code>null</code> if no final status is expected.
     */
    public static ServiceProviderImplementation faultMessage(final String content, final MessageChecks statusChecks) {
        return new ServiceProviderImplementation() {
            @Override
            public Message provides(final RequestMessage request) throws Exception {
                return new FaultToConsumerMessage(request, content);
            }

            @Override
            public void handleStatus(final StatusMessage status) throws Exception {
                assert statusChecks != null;
                statusChecks.checks(status);
            }

            @Override
            public boolean statusExpected() {
                return statusChecks != null;
            }
        };
    }

    /**
     * A service that returns a ResponseMessage with the given content as fault with {@link ExchangeStatus#DONE} as
     * expected final status.
     */
    public static ServiceProviderImplementation faultMessage(final String content) {
        return faultMessage(content, MessageChecks.status(ExchangeStatus.DONE));
    }

    public static ServiceProviderImplementation statusMessage(final ExchangeStatus status) {
        return new ServiceProviderImplementation() {
            @Override
            public Message provides(final RequestMessage request) throws Exception {
                return new StatusToConsumerMessage(request, status);
            }

            @Override
            public boolean statusExpected() {
                return false;
            }
        };
    }

    public static ServiceProviderImplementation errorMessage(final Exception error) {
        return new ServiceProviderImplementation() {
            @Override
            public Message provides(final RequestMessage request) throws Exception {
                return new StatusToConsumerMessage(request, error);
            }

            @Override
            public boolean statusExpected() {
                return false;
            }
        };
    }
}